/*
 * Icons.h - Threaded shell icon retrieval.
 */

#pragma once

#define USE_CRT_THREADING

#include "std.h"
#include "str.h"

class AsyncShellIcon : public RefCounted
{
public:
					AsyncShellIcon();
					~AsyncShellIcon();

	static UINT const c_wmIconAvailable;

	void			Enqueue( HWND hwndNotify, char const* pszFile, ULONG ulShgfi );

	void			Cancel();

	bool			IsComplete() const { return m_fComplete; }
	int				RetrieveIconIndex();

	static void		Uninitialize( bool fWeakWait = false );

private:
	void			AsyncGetIcon();
	void			Finish();

private:
	HWND			m_hwndNotify;
	Str				m_sFile;
	ULONG			m_ulShgfi;
	bool			m_fQueued;
	bool			m_fComplete;
	int				m_iIcon;

#ifdef DEBUG
	bool			m_fZombie;
#endif

	AsyncShellIcon* m_pNext;
	AsyncShellIcon* m_pPrev;

	static AsyncShellIcon* s_pHead;
	static AsyncShellIcon* s_pTail;

#ifdef USE_CRT_THREADING
	static unsigned __stdcall ThreadProc( void* pv );
#else
	static DWORD WINAPI ThreadProc( void* pv );
#endif
};

