#include <windows.h>
#include "std.h"

#ifdef DEBUG

inline void
DbgTraceRaw( char const* psz )
{
	OutputDebugString( psz );
}

static void
DbgTraceV( bool fNewLine, char const* pszFormat, char* pvargs )
{
	char szBuffer[2048];

	int const cch = wvsprintfA( szBuffer, pszFormat, pvargs );
	AssertSz( cch < DimensionOf( szBuffer ), "DbgTraceV overran buffer!" );
	if( fNewLine )
	{
		int const ichNewline = Min( size_t( cch ), DimensionOf( szBuffer ) - 3 );
		lstrcpy( szBuffer + ichNewline, "\r\n" );
	}
	DbgTraceRaw( szBuffer );
}

void
DbgTrace_( char const* pszFormat, ... )
{
	va_list args;
	va_start( args, pszFormat );
	DbgTraceV( false, pszFormat, args );
	va_end( args );
}

void
DbgTrace( char const* pszFormat, ... )
{
	va_list args;
	va_start( args, pszFormat );
	DbgTraceV( true, pszFormat, args );
	va_end( args );
}

void
DbgAssert( char const* pszFileName, int iLine, char const* pszFmt, ... )
{
	static bool s_fInDbgAssert = false;
	va_list arg;
	char sz[2048];
	int ival;
	char *pch = sz;
	char *pchEnd;
	MSG msgT;

	pch += wsprintfA( pch, "%s @ line %d", pszFileName, iLine );

	if( pszFmt )
	{
		*pch++ = ':';
		*pch++ = ' ';
		*pch++ = '"';

		va_start( arg, pszFmt );
		pch += wvsprintfA( pch, pszFmt, arg );
		va_end( arg );

		// Remove trailing newlines...
		while( *( pch-1 ) == '\n' )
			--pch;

		*pch++ = '"';
		*pch++ = '\r';
		*pch++ = '\n';
	}
	else
	{
		*pch++ = ' ';
	}

	if( s_fInDbgAssert )
	{
		*pch = 0;
		Trace( "Arrgg! Recursive assert: %s", sz );
		return;
	}

	s_fInDbgAssert = true;

	*pch = '\0';
	pchEnd = pch;

	Trace( "ASSERT:" );
	Trace( sz );

	// If there a WM_QUIT message in the queue, remote it and add it again
	// after the message box is dismissed.

	BOOL const fQuit = ::PeekMessage( &msgT, 0, WM_QUIT, WM_QUIT, PM_REMOVE );
	ival = MessageBoxA( 0, sz, "Assertion", MB_ICONEXCLAMATION|MB_ABORTRETRYIGNORE|MB_DEFBUTTON3 );
	if( fQuit )
		::PostQuitMessage( msgT.wParam );

	switch( ival )
	{
	case 0:
		Trace( "Failed to create message box on assert." );
		// Fall through...
	case IDRETRY:
		s_fInDbgAssert = false;
		DebugBreak();
		return;
	case IDABORT:
		ExitProcess( 1 );
		return;
	default:
		Trace( "Unexpected return value from assertion message box %d.", ival );
		DebugBreak();
		// Fall through...
	case IDIGNORE:
		break;
	}

	s_fInDbgAssert = false;
}

#endif

