#include "projutil.h"
#include <richedit.h>
#include <richole.h>
#include <zmouse.h>
#include "std.h"
#include "util.h"
#include "dlg.h"
#include "res.h"

/*
 * IRichEditOleCallback helpers.
 */

class RichEditHelper : public RefCounted, public IRichEditOleCallback
{
	typedef RefCounted base;

public:
					RichEditHelper( HWND hwnd, unsigned cchLimit = 0, DWORD dwEnmMask = 0 );
					~RichEditHelper();

	DeclareIUnknownMembers( IMPL );
	DeclareIRichEditOleCallbackMembers( IMPL );

	bool			IsReadOnly() const { return !!( ::GetWindowLong( m_hwnd, GWL_STYLE ) & ES_READONLY ); }
	DWORD			GetEventMask() const { return ::SendMessage( m_hwnd, EM_GETEVENTMASK, 0, 0 ); }
	DWORD			SetEventMask( DWORD dwEventMask ) { return ::SendMessage( m_hwnd, EM_SETEVENTMASK, 0, dwEventMask ); }

private:
#ifndef STRICT		// See help on CallWindowProc.
#error Must compile with STRICT for CallWindowProc to work right.
#endif
	HWND const		m_hwnd;
};

RichEditHelper::RichEditHelper( HWND hwnd, unsigned cchLimit, DWORD dwEnmMask )
	: m_hwnd( hwnd )
{
	int const c_enmBasic = ENM_CHANGE | ENM_SELCHANGE | ENM_MOUSEEVENTS;
	//int const c_enmRich = ENM_CHANGE | ENM_SELCHANGE | ENM_MOUSEEVENTS | ENM_LINK;
	int const c_cxLeftMargin = 1;
	int const c_cxRightMargin = 1;

	::SendMessage( m_hwnd, EM_SETTEXTMODE, TM_MULTILEVELUNDO|TM_PLAINTEXT|TM_SINGLECODEPAGE, 0 );
	::SendMessage( m_hwnd, EM_SETEVENTMASK, 0, c_enmBasic | dwEnmMask );
	::SendMessage( m_hwnd, EM_SETMARGINS, EC_LEFTMARGIN|EC_RIGHTMARGIN, MAKELONG( c_cxLeftMargin, c_cxRightMargin ) );
	::SendMessage( m_hwnd, EM_SETOLECALLBACK, 0, LONG( static_cast<IRichEditOleCallback*>( this ) ) );
	if( cchLimit )
		::SendMessage( m_hwnd, EM_LIMITTEXT, cchLimit, 0 );

	if( ::GetWindowLong( m_hwnd, GWL_EXSTYLE ) & WS_EX_CLIENTEDGE )
		ThemeSubclassEdit( m_hwnd );
}

RichEditHelper::~RichEditHelper()
{
	if( m_hwnd )
		::SendMessage( m_hwnd, EM_SETOLECALLBACK, 0, 0 );
}

STDMETHODIMP
RichEditHelper::QueryInterface( REFCLSID clsid, void** ppv )
{
	if( clsid == IID_IUnknown || clsid == IID_IRichEditOleCallback )
		*ppv = static_cast<IRichEditOleCallback*>( this );
	else
		return E_NOINTERFACE;
	AddRef();
	return S_OK;
}

STDMETHODIMP_( ULONG )
RichEditHelper::AddRef()
{
	return base::AddRef();
}

STDMETHODIMP_( ULONG )
RichEditHelper::Release()
{
	return base::Release();
}

STDMETHODIMP
RichEditHelper::GetNewStorage( IStorage** )
{
	return E_NOTIMPL;
}

STDMETHODIMP
RichEditHelper::GetInPlaceContext( IOleInPlaceFrame**, IOleInPlaceUIWindow**, OLEINPLACEFRAMEINFO* )
{
	return E_NOTIMPL;
}

STDMETHODIMP
RichEditHelper::ShowContainerUI( BOOL fShow )
{
	return E_NOTIMPL;
}

STDMETHODIMP
RichEditHelper::QueryInsertObject( CLSID* pclsid, IStorage* pstg, LONG cp )
{
	return S_OK;
}

STDMETHODIMP
RichEditHelper::DeleteObject( IOleObject* poo )
{
	return E_NOTIMPL;
}

STDMETHODIMP
RichEditHelper::QueryAcceptData( IDataObject* pdo, CLIPFORMAT* pcfFormat, DWORD reco, BOOL fReally, HGLOBAL hMetaPict )
{
	if( IsReadOnly() )
		return E_FAIL;

	if( !*pcfFormat )
	{
		// Aak for CF_TEXT.

		FORMATETC fe = { 0 };
		fe.tymed = TYMED_ISTORAGE;
		fe.cfFormat = CF_TEXT;
		fe.dwAspect = DVASPECT_CONTENT;
		fe.lindex = -1;

		if( pdo->QueryGetData( &fe ) == S_OK )
			*pcfFormat = fe.cfFormat;
	}

	if( *pcfFormat != CF_TEXT )
		return E_FAIL;

	return S_OK;
}

STDMETHODIMP
RichEditHelper::ContextSensitiveHelp( BOOL fEnterMode )
{
	return E_NOTIMPL;
}

STDMETHODIMP
RichEditHelper::GetClipboardData( CHARRANGE* pchrg, DWORD reco, IDataObject** ppdo )
{
	return E_NOTIMPL;
}

STDMETHODIMP
RichEditHelper::GetDragDropEffect( BOOL fSource, DWORD grfKeyState, LPDWORD pdwEffect )
{
	return E_NOTIMPL;
}

STDMETHODIMP
RichEditHelper::GetContextMenu( WORD seltype, IOleObject* poo, CHARRANGE* pchrg, HMENU* phmenu )
{
	*phmenu = 0;

	if( seltype == GCM_RIGHTMOUSEDROP )
		return S_OK;

	HMENU const hmenu = LoadMenuU( IDR_RICHEDIT_CTX );
	*phmenu = GetSubMenu( hmenu, 0 );
	RemoveMenu( hmenu, 0, MF_BYPOSITION );
	DestroyMenu( hmenu );

	return S_OK;
}

/*
 * DlgBase.
 */

UINT const c_wmMouseWheel = RegisterWindowMessage( MSH_MOUSEWHEEL );

DlgBase::DlgBase()
	: m_idDlg( 0 )
	, m_hwnd( 0 )
{
}

DlgBase::~DlgBase()
{
}

INT_PTR CALLBACK
DlgBase::StaticDlgProc( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	if( msg == WM_INITDIALOG )
	{
		Assert( lParam );
		SetWindowLongPtr( hwnd, DWLP_USER, lParam );
		reinterpret_cast<DlgBase*>( lParam )->m_hwnd = hwnd;
	}

	DlgBase* const pThis = reinterpret_cast<DlgBase*>( GetWindowLongPtr( hwnd, DWLP_USER ) );

	if( !pThis )
	{
		Trace( "DlgBase::StaticDlgProc unable to handle msg = 0x%08x sent to hwnd = 0x%08x (pThis == 0x%08x)", msg, hwnd, pThis );
#ifdef DEBUG
		if( msg == WM_MEASUREITEM )
		{
			Trace( "** NOTE: For controls with the OWNERDRAWFIXED style, Windows sends WM_MEASUREITEM before it sends WM_INITDIALOG, and DlgBase does not support this.\r\n"
				   "** Instead of WM_MEASUREITEM, use LB_SETITEMHEIGHT and LB_SETCOLUMNWIDTH, etc." );
		}
#endif
		return false;
	}

	Assert( pThis->m_hwnd == hwnd );

	INT_PTR const fResult = pThis->DlgProc( msg, wParam, lParam );

	if( msg == WM_NCDESTROY )
	    pThis->m_hwnd = 0;

	// Must return actual result in order for things like WM_CTLCOLORLISTBOX
	// to work.
	return fResult;
}

static HWND
__GetSafeOwner( HWND hwndParent, HWND* phwndTop )
{
	// Get window to start with.

	//$ review: (chrisant) If the hwndParent!=NULL requirement is ever to be
	// relaxed, be sure to consult the MFC implementation again.

	Assert( hwndParent );

	HWND hwnd = hwndParent;

	// A popup window cannot be owned by a child window.

	while( !!hwnd && ( GetWindowStyle( hwnd ) & WS_CHILD ) )
		hwnd = GetParent( hwnd );

	Assert( hwnd );

	// Determine toplevel window to disable as well.

	HWND hwndTop = hwnd;
	HWND hwndWalk = hwnd;
	while( hwndWalk )
	{
		hwndTop = hwndWalk;
		hwndWalk = GetParent( hwndWalk );
	}

	// Disable and store top level parent window if specified.

	if( phwndTop )
	{
		*phwndTop = 0;

		if( !!hwndTop && IsWindowEnabled( hwndTop ) && hwndTop != hwnd )
		{
			*phwndTop = hwndTop;
			EnableWindow( hwndTop, false );
		}
	}

	// Return the owner.

	Assert( hwnd );
	Assert( !( GetWindowStyle( hwnd ) & WS_CHILD ) );

	return hwnd;
}

INT_PTR
DlgBase::DoModal( HWND _hwndParent, UINT idDlg, UINT idrAccel )
{
	Assert( idDlg );
	Assert( _hwndParent );

	Assert( !m_idDlg );
	Assert( !m_hwnd );

	m_idDlg = idDlg;

#ifndef NEED_DIALOG_ACCELERATORS

	Assert( !idrAccel );
	INT_PTR const nRet = DialogBoxParamU( idDlg, _hwndParent, StaticDlgProc, LPARAM( this ) );

#else

	HWND hwndTop;
	HWND const hwndParent = __GetSafeOwner( _hwndParent, &hwndTop );
	bool fEnableParent = false;

	if( !!hwndParent && IsWindowEnabled( hwndParent ) )
	{
		EnableWindow( hwndParent, false );
		fEnableParent = true;
	}

	// Create modeless dialog.

	HWND const hwndFocus = GetFocus();

	m_haccel = idrAccel ? LoadAcceleratorsU( idrAccel ) : 0;
	m_fContinueModal = true;

	if( CreateDialogParamU( idDlg, hwndParent, StaticDlgProc, LPARAM( this ) ) )
	{
		Assert( IsWindow( m_hwnd ) );

		// Pump messages until the modal state is done.

		bool fShowOnIdle = !( GetWindowStyle( m_hwnd ) & WS_VISIBLE );

		while( m_fContinueModal )
		{
			if( fShowOnIdle )
			{
				MSG msg;

				if( !PeekMessage( &msg, 0, 0, 0, PM_NOREMOVE ) )
				{
					fShowOnIdle = false;
					ShowWindow( m_hwnd, SW_SHOW );
				}
				else
				{
					// If the dialog was not displayed yet and the current
					// message is a mouse message for a disabled window get
					// rid of it in order to be able to display the dialog
					// successfully.

					if( ( msg.message >= WM_MOUSEFIRST && msg.message <= WM_MOUSELAST ) &&
						!( IsWindowEnabled( msg.hwnd ) ) )
					{
						PeekMessage( &msg, 0, 0, 0, PM_REMOVE );
						continue;
					}
				}
			}

			// Pump message, but quit on WM_QUIT.

			MSG msg;

#if 0
			if( !PeekMessage( &msg, 0, 0, 0, PM_NOREMOVE ) )
			{
				//$ FUTURE: (chrisant) Could implement idle callback here.
			}
#endif

			if( GetMessage( &msg, NULL, 0, 0 ) )
			{
				if( msg.message == c_wmMouseWheel || msg.message == WM_MOUSEWHEEL )
				{
					// Hover scrolling for windows in our thread.
					HWND const hwndPoint = WindowFromPoint( msg.pt );
					if( !IsChild( hwndPoint, ::GetFocus() ) &&
						GetWindowThreadProcessId( hwndPoint, NULL ) == GetCurrentThreadId() )
					{
						msg.hwnd = hwndPoint;
					}
				}

				if( !m_haccel || !TranslateAccelerator( m_hwnd, m_haccel, &msg ) )
				{
					if( !IsDialogMessage( m_hwnd, &msg ) )
					{
						TranslateMessage( &msg );
						DispatchMessage( &msg );
					}
				}
			}
			else
			{
				PostQuitMessage( msg.wParam );
				EndDialog( -1 );
			}
		}

		m_fContinueModal = false;

		// Hide the window before enabling the parent, etc.

		if( m_hwnd )
			SetWindowPos( m_hwnd, 0, 0, 0, 0, 0, SWP_HIDEWINDOW|SWP_NOSIZE|SWP_NOMOVE|SWP_NOACTIVATE|SWP_NOZORDER );
	}

	if( fEnableParent )
		EnableWindow( hwndParent, true );

	if( !!hwndParent && GetActiveWindow() == m_hwnd )
		SetActiveWindow( hwndParent );

	DestroyWindow( m_hwnd );

	// Re-enable windows.

	if( !!hwndTop && IsWindow( hwndTop ) )
		EnableWindow( hwndTop, true );

	if( hwndFocus )
	{
		HWND hwndFocusParent = GetParent( hwndFocus );

		if( hwndFocusParent )
			::SendMessage( hwndFocusParent, WM_NEXTDLGCTL, WPARAM( hwndFocus ), 0 );
		else
			SetFocus( hwndFocus );
	}

	INT_PTR const nRet = m_nResult;

#endif

	m_idDlg = 0;
	m_rgReleaseOnClose.Clear();

	Assert( !m_hwnd );

	return nRet;
}

BOOL
DlgBase::EndDialog( INT_PTR nResult )
{
	Assert( m_hwnd );
#ifndef NEED_DIALOG_ACCELERATORS
	return ::EndDialog( m_hwnd, nResult );
#else
	m_nResult = nResult;
	if( m_fContinueModal )
	{
		m_fContinueModal = false;
		::PostMessage( m_hwnd, WM_NULL, 0, 0 );
	}
	return true;
#endif
}

LRESULT
DlgBase::SendMessage( UINT msg, WPARAM wParam, LPARAM lParam )
{
	Assert( m_hwnd );
	return ::SendMessage( m_hwnd, msg, wParam, lParam );
}

LRESULT
DlgBase::SendDlgItemMessage( UINT idc, UINT msg, WPARAM wParam, LPARAM lParam )
{
	Assert( m_hwnd );
	return ::SendDlgItemMessage( m_hwnd, idc, msg, wParam, lParam );
}

HWND
DlgBase::GetDlgItem( UINT idc )
{
	Assert( m_hwnd );
	return ::GetDlgItem( m_hwnd, idc );
}

/**
 * SetDlgItemFocus - Properly set focus to a dialog control.
 *
 * @param idc					Control ID to which to set focus.
 * @param fNext					When <i>idc</i> is zero, sets focus to next or
 *								previous control depending on <i>fNext</i>.
 */
void
DlgBase::SetDlgItemFocus( UINT idc, bool fNext )
{
	SendMessage( WM_NEXTDLGCTL, idc ? LPARAM( GetDlgItem( idc ) ) : 0, idc != 0 );
}

/**
 * DisableControl - Safely disable a control in the dialog.
 *
 * If the control has focus then focus will altered and if it is the default
 * control then the default control will be altered.
 *
 * @param idc					Control to disable.
 * @param idcMaybeNewFocus		Control to receive focus if <i>idc</i> has
 *								focus. When zero, focus goes to the next
 *								control in the tab order.
 * @param idcMaybeNewDefault	Control to become the default control if
 *								<i>idc</i> is the default control.
 *
 * @return BOOL					TRUE if focus has been modified, or FALSE if
 *								focus has not been modified.
 *
 * @note						If this method is called from a WM_INITDIALOG
 *								handler then the return code can be used as
 *								the return from that handler.
 */
BOOL
DlgBase::DisableControl( UINT idc, UINT idcMaybeNewFocus, UINT idcMaybeNewDefault )
{
	Assert( idc != idcMaybeNewFocus );
	Assert( idc != idcMaybeNewDefault );

	HWND const hwndCtl = GetDlgItem( idc );
	BOOL fRc = true;

	// First see if the control has focus.

	if( GetFocus() == hwndCtl )
	{
		// Move focus to the other control.  If idcMaybeNewFocus is zero then
		// focus is moved to the next control in the tab sequence.

		fRc = false;
		SetDlgItemFocus( idcMaybeNewFocus );
	}

	// Now check that the control is not the default control.

	DWORD dwStyle = GetWindowLong( hwndCtl, GWL_STYLE );

	if( dwStyle & BS_DEFPUSHBUTTON )
	{
		// Apparently the caller is attempting to disable the default button.
		// Move the default to the suggested target.

		DWORD dwTargetStyle;

		if( idcMaybeNewDefault )
			dwTargetStyle = GetWindowLong( GetDlgItem( idcMaybeNewDefault ), GWL_STYLE );

		dwStyle &= ~BS_DEFPUSHBUTTON;
		dwStyle |= BS_PUSHBUTTON;

		// Turn off the default button.

		SendDlgItemMessage( idc, BM_SETSTYLE, dwStyle, true );

		if( idcMaybeNewDefault )
		{
			dwTargetStyle |= BS_DEFPUSHBUTTON;
			SendDlgItemMessage( idcMaybeNewDefault, BM_SETSTYLE, dwTargetStyle, true );
		}
	}

	// Now it's safe to disable the control.

	EnableWindow( hwndCtl, false );

	return fRc;
}

/**
 * EnableControl - Enable a control in the dialog.
 *
 * @param idc					Control to enable.
 *
 * @return BOOL					TRUE if the window was previously disabled,
 *								otherwise FALSE.
 */
BOOL
DlgBase::EnableControl( UINT idc )
{
	return EnableWindow( GetDlgItem( idc ), true );
}

void
DlgBase::EnableControl( UINT idc, bool fEnable )
{
	HWND const hwndCtl = GetDlgItem( idc );

	if( fEnable )
	{
		ShowWindow( hwndCtl, SW_SHOW );
		EnableControl( idc );
	}
	else
	{
		DisableControl( idc );
		ShowWindow( hwndCtl, SW_HIDE );
	}
}

void
DlgBase::InitRichEdit( UINT idc, unsigned cchLimit, DWORD dwEnmMask )
{
	SPIUnknown* pspunk;
	m_rgReleaseOnClose.Append( &pspunk );
	pspunk->Assign( new RichEditHelper( GetDlgItem( idc ), cchLimit, dwEnmMask ) );
}

bool
DlgBase::HandleEditCommand( HWND hwndControl, UINT idm )
{
	if( !hwndControl )
		hwndControl = GetFocus();

	if( !hwndControl )
		return false;

	switch( idm )
	{
	case IDM_EDITUNDO:		::SendMessage( hwndControl, EM_UNDO, 0, 0 ); break;
	case IDM_EDITCUT:		::SendMessage( hwndControl, WM_CUT, 0, 0 ); break;
	case IDM_EDITCOPY:		::SendMessage( hwndControl, WM_COPY, 0, 0 ); break;
	case IDM_EDITPASTE:		::SendMessage( hwndControl, WM_PASTE, 0, 0 ); break;
	case IDM_EDITCLEAR:		::SendMessage( hwndControl, WM_CLEAR, 0, 0 ); break;
	case IDM_EDITSELECTALL:	::SendMessage( hwndControl, EM_SETSEL, 0, -1 ); break;
	default:
		return false;
	}

	return true;
}

