/*
 * Str.h - Simplistic string class.
 */

#pragma once

//#define USE_LFH_FOR_STR

int sprintf_s( char* pszOut, size_t cchMax, char const* pszFormat, ... );
int vsprintf_s( char* pszOut, size_t cchMax, char const* pszFormat, va_list args );

class Str
{
public:
					Str() { Init(); }
					~Str();

	static void		InitializeHeap();
	static void		UninitializeHeap();

	char const*		Text() const { return m_psz; }
	size_t			Length() const { return m_cchLen; }
	size_t			AllocSize() const { return m_cchAlloc; }

	void			Clear();
	void			Free();
	void			Init();

	bool			Equal( char const* psz ) const { return !lstrcmp( psz, Text() ); }
	bool			Equal( Str const* s ) const;
	bool			Equal( Str const& s ) const { return Equal( &s ); }
	bool			EqualI( char const* psz ) const { return !lstrcmpi( psz, Text() ); }
	bool			EqualI( Str const* s ) const;
	bool			EqualI( Str const& s ) const { return EqualI( &s ); }

	void			SetEnd( char const* pszEnd );
	void			SetLength( size_t cch );

	void			Set( char const* psz, size_t cch = -1 );
	void			Set( Str const& s ) { Set( s.Text(), s.Length() ); }

	char*			Alloc( size_t cch );

	void			Append( char ch );
	void			Append( char const* psz, size_t cch = -1 );
	void			Append( Str const& s ) { Append( s.Text(), s.Length() ); }

	void			PrintfV( char const* pszFormat, va_list args );
	void			Printf( char const* pszFormat, ... );
#if 0
	void			Printf( UINT idsFormat, ... );
	size_t			LoadString( UINT ids );
#endif
	size_t			GetWindowText( HWND hwnd );

	void			TrimLeft();
	void			TrimRight();
	void			Trim();

#if 0
	bool			GetToken( Str& out );
#endif

#if 0
	void			Attach( char* psz, size_t cchLen, size_t cchAlloc );
	void			Swap( Str& other );
#endif

private:
	char*			AllocInternal( size_t cch );
	char*			MemReAlloc( char* pszExisting, size_t cchNewSize );
	void			MemFree( char* pszExisting );

private:
	char*			m_psz;
	size_t			m_cchLen;
	size_t			m_cchAlloc;

	static char const* const s_pszEmpty;
#ifdef USE_LFH_FOR_STR
	static HANDLE s_hHeap;
#endif
};

